/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.TryCatchBlockNode;


/**
 * Generic exception handler.
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public final class CodeExceptionGen implements InstructionTargeter {

	/**
	 * Stateful adapter.
	 */
	@WrapperMember
	protected ObjectType type = null;
	
	/**
	 * The list that holds this exception handler.
	 */
	@WrapperMember
	public InstructionList list = null;
	
	/**
	 * Wrappee field.
	 */
	@WrapperMember
	private TryCatchBlockNode node = null; 
	
	/**
	 * Wrapper constructor.
	 *
	 * @param node
	 */
	@WrapperMember
	public CodeExceptionGen(TryCatchBlockNode node) {
		setNode(node);
		this.type = new ObjectType(node.type);
	}

	/**
	 * Sets the class node wrappee, registering in the identity map.
	 *
	 * @param node
	 */
	@WrapperMember
	protected TryCatchBlockNode setNode(TryCatchBlockNode node) {
		this.node = node;
		WrapperRecycler.add(this, node);
		return node;
	}
	
	/**
	 * Getter allows for wrappee refinement.
	 *
	 * @return
	 */
	@WrapperMember
	protected TryCatchBlockNode getNode() {
		return node;
	}
	
	/**
	 * Creates a new code exception gen with all data.
	 *
	 * @param start_pc
	 * @param end_pc
	 * @param handler_pc
	 * @param catch_type
	 */
	public CodeExceptionGen(InstructionHandle start_pc, InstructionHandle end_pc, InstructionHandle handler_pc, ObjectType catch_type) {
		
		assert(start_pc.list == end_pc.list);
		assert(start_pc.list == handler_pc.list);
		
		// Note that we use next label for end_pc because BCEL treats the exception region inclusively.
		// But deal with the case that end_pc is the last instruction on the list.
		this.setNode(new TryCatchBlockNode(
				start_pc.ensureLabel(), 
				end_pc.ensureNextLabel(), 
				handler_pc.ensureLabel(),
				catch_type.getType().getInternalName()));
		
		// Mark this code exception gen as the targeter of those handlers
		start_pc.addTargeter(this);
		end_pc.addTargeter(this);
		handler_pc.addTargeter(this);
		
		// Need to store the type to obey object identity
		this.type = catch_type;
		
		// All instruction handles are assumed to come from the same list
		this.list = start_pc.list;
	}

	public InstructionHandle getStartPC() {
		return list.getHandle(node.start);
	}

	public InstructionHandle getEndPC() {
		return list.getHandle(node.end).getPrev();
	}

	public InstructionHandle getHandlerPC() {
		return list.getHandle(node.handler);
	}

	public ObjectType getCatchType() {
		return type;
	}
	
	@WrapperMember
	protected void replace(InstructionHandle ih) {
		InstructionHandle next = ih.getNext();
		
		if (ih.hasLabel()) {
			LabelNode n = ih.ensureLabel();
			if (node.start == n)
				node.start = next.ensureLabel();
			else if (node.handler == n)
				node.handler = next.ensureLabel();
		} else {
			assert(ih.hasNextLabel());
			assert(node.end == ih.ensureNextLabel());
			node.end = next.ensureNextLabel();
		}
		next.addTargeter(this);
		ih.removeTargeter(this);
	}
	
	/**
	 * Returns all labels that are targets of this instruction.
	 *
	 * @return
	 */
	@WrapperMember
	public LabelNode[] getTargetLabels() {
		TryCatchBlockNode n = (TryCatchBlockNode) getNode();
		return new LabelNode[] {n.start, n.end, n.handler};
	}
}

