/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.IincInsnNode;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.LineNumberNode;
import org.objectweb.asm.tree.LookupSwitchInsnNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MultiANewArrayInsnNode;
import org.objectweb.asm.tree.TableSwitchInsnNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;


/**
 * 
 * TODO Comment!!
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public class InstructionHandle {

	
	private static Queue<InstructionHandle> pool = new LinkedList<InstructionHandle>();
	
	/**
	 * Wrappee field.
	 */
	@WrapperMember
	private AbstractInsnNode node = null;
	
	/**
	 * The instruction being handled.
	 * 
	 * Stateful adapter. Essentially a re-implementation.
	 */
	@WrapperMember
	protected Instruction instruction = null;
	
	/**
	 * The list to which this instruction was appended, if any.
	 */
	@WrapperMember
	public InstructionList list = null;
	
	/**
	 * The reverse pointer on a target relation.
	 * 
	 * That is, this set keeps other instructions in the list that target this instruction.
	 */
	@WrapperMember
	protected Set<InstructionTargeter> targeters = null;
	
	/**
	 * The offset of the handle in the list.
	 */
	@WrapperMember
	private int position = -1;
	
	/**
	 * Wrapper constructor.
	 *
	 * @param node
	 */
	@WrapperMember
	public InstructionHandle(AbstractInsnNode node) {
		setNode(node);
	}
	
	static InstructionHandle createHandle(AbstractInsnNode node) {
		if (! pool.isEmpty()) {
			InstructionHandle handle = pool.poll();
			handle.setNode(node);
			return handle;
		}
		return new InstructionHandle(node);
	}
	
	public void dispose() {
		//WrapperRecycler.remove(instruction, node);
		WrapperRecycler.remove(this, node);
		this.instruction.removeHandle(this);
		this.node = null;
		this.instruction = null;
		this.list = null;
		this.targeters = null;
		this.position = -1;
		pool.add(this);
	}
	
	/**
	 * Sets the class node wrappee, registering in the identity map.
	 *
	 * @param node
	 */
	@WrapperMember
	protected AbstractInsnNode setNode(AbstractInsnNode node) {
		this.node = node;
		WrapperRecycler.add(this, node);
		return node;
	}
	
	/**
	 * Getter allows for wrappee refinement.
	 *
	 * @return
	 */
	@WrapperMember
	protected AbstractInsnNode getNode() {
		return node;
	}
		
	@WrapperMember
	public LineNumberNode getLineNumber() {
		return list.getLineNumber(this);
	}
	
	/**
	 * Creates an empty instruction handle.
	 */
	@WrapperMember
	protected InstructionHandle() {
	}

	@WrapperMember
	InstructionHandle setInstruction(Instruction i) {
		this.instruction = i;
		// Set the relation between the instruction and this handler's node
		WrapperRecycler.add(i, this.getNode());
		return this;
	}
	
	/**
	 * 
	 * TODO Comment!!
	 *
	 * @return
	 */
	public int getPosition() {
		return position;
	}
	
	/**
	 * Sets this handler's position.
	 *
	 * @param pos
	 */
	void setPosition(int pos) {
        position = pos;
    }
	
	/**
	 * Gets the instruction wrapped by this handle.
	 *
	 * @return
	 */
	public final Instruction getInstruction() {
		if (null == instruction) {
			instruction = Instruction.createInstruction(this.getNode());
			instruction.addHandle(this);
		}
		return instruction;
	}
	
	/**
	 * Gets the instruction, but only if it is a targeter.
	 *
	 * @return
	 */
	@WrapperMember
	public final InstructionTargeter getInstructionTargeter() {
		Instruction i = getInstruction();
		if (i instanceof InstructionTargeter)
			return (InstructionTargeter) i;
		return null;
	}

	/**
	 * Denotes that this targeter points to this handle.
	 *
	 * @param t
	 */
	public void addTargeter(InstructionTargeter t) {
		if (null == targeters)
			targeters = new HashSet<InstructionTargeter>();
		targeters.add(t);
	}
	
	@WrapperMember
	public boolean isTargeter(InstructionTargeter t) {
		if (null == targeters)
			return false;
		return targeters.contains(t);
	}
	
	
	/**
	 * Removes this targeter from the set.
	 *
	 * @param t
	 */
	public void removeTargeter(InstructionTargeter t) {
		if (null != targeters)
			targeters.remove(t);
	}
	
	/**
	 * Gets the targeters of this instruction.
	 *
	 * @return
	 */
	public InstructionTargeter[] getTargeters() {
		 if (null == targeters || targeters.size() == 0) 
			 return null;
		 
        InstructionTargeter[] array = new InstructionTargeter[targeters.size()];
        targeters.toArray(array);
        return array;
	}

	/**
	 * Checks if there are other instructions in the list that point to this instruction.
	 *
	 * @return
	 */
	public boolean hasTargeters() {
        return targeters != null && targeters.size() > 0;
	}

	/**
	 * Gets a handle to the next instruction in the list, or null if this is the last instruction.
	 *
	 * @return
	 */
	public final InstructionHandle getNext() {
		return list.getNext(this.getNode());
	}

	/**
	 * Gets a handle to the previous instruction in the list, or null if this is the first instruction. 
	 *
	 * @return
	 */
	public final InstructionHandle getPrev() {
		return list.getPrev(this.getNode());
	}
	
	
	/**
	 * Ensures that this handle has a label, and returns the label.
	 *
	 * @return
	 */
	@WrapperMember
	public LabelNode ensureLabel() {
		return list.ensureLabel(this);
	}
	
	/**
	 * Checks whether the handle as a label already.
	 *
	 * @return
	 */
	@WrapperMember
	public boolean hasLabel() {
		return list.hasLabel(this);
	}
	
	/**
	 * Checks whether the handle as a next label already.
	 *
	 * @return
	 */
	@WrapperMember
	public boolean hasNextLabel() {
		return list.hasNextLabel(this);
	}
	
	/**
	 * Ensures that this handle has a next label, and returns the label.
	 *
	 * @return
	 */
	@WrapperMember
	public LabelNode ensureNextLabel() {
		return list.ensureNextLabel(this);
	}
	
	/**
	 * Computes how many bytes the instruction represented by this node uses.
	 * 
	 * This code is based on the offsets used by MethodWriter.
	 *
	 * @return
	 */
	@WrapperMember
	public static int getByteCount(AbstractInsnNode node) {
		
		if (node instanceof InsnNode)
			return 1;
		
		if (node instanceof TypeInsnNode)
			return 3;
		
		if (node instanceof FieldInsnNode)
			return 3;
		
		if (node instanceof MultiANewArrayInsnNode)
			return 4;
		
		int opcode = node.getOpcode();
		if (node instanceof IntInsnNode)
			return (opcode == Opcodes.SIPUSH) ? 3 : 2;
		
		if (node instanceof VarInsnNode) {
			int var = ((VarInsnNode) node).var;
			return (var < 4 && opcode != Opcodes.RET) ?	1 :
				   (var >= 256) ? 4 :
				    2;
		}
		
		if (node instanceof MethodInsnNode)
			return (opcode == Opcodes.INVOKEINTERFACE) ? 5 : 
				   (opcode == Opcodes.INVOKEDYNAMIC) ? 4 : 
					3;
		
		if (node instanceof LdcInsnNode) {
			Object cst = ((LdcInsnNode) node).cst;
			// TODO - missing the case when this constant is in an index larger
			// than 256, in which case we should return 3 instead of 2...
			return (cst instanceof Long || cst instanceof Double) ? 3 : 2;
		}
		
		if (node instanceof IincInsnNode) {
			int incr = ((IincInsnNode) node).incr;
			int var = ((IincInsnNode) node).var;
			return ((var > 255) || (incr > 127) || (incr < -128)) ? 6 : 3;
		}
		
		// TODO - not sure about these instructions
		if (node instanceof JumpInsnNode) {
			return 3;
		}
		if (node instanceof TableSwitchInsnNode) {
			return 9;
		}
		if (node instanceof LookupSwitchInsnNode) {
			return (((LookupSwitchInsnNode) node).labels.size() + 1) * 8 + 3;
		}
		
		throw new RuntimeException("Missing case in InstructionHandle.getByteCount()");
	}
}
