/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.FrameNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LineNumberNode;


/**
 * A list of instructions.
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public class InstructionList {

	/**
	 * Wrappee field.
	 */
	@WrapperMember
	protected InsnList list = null;
	
	/**
	 * Wrapper constructor.
	 *
	 * @param list
	 */
	@WrapperMember
	InstructionList(InsnList list) {
		setList(list);
	}
	
	/**
	 * Sets the wrappee, registering in the identity map.
	 * 
	 * @param list
	 */
	@WrapperMember
	protected void setList(InsnList list) {
		this.list = list;
		WrapperRecycler.add(this, list);
	}
	
	/**
	 * Initializes the list with an instruction.
	 *
	 * @param i
	 */
	public InstructionList(Instruction i) {
		this();
		this.append(i);
	}

	/**
	 * Creates an empty instruction list.
	 */
	public InstructionList() {
		setList(new InsnList());
	}

	/**
	 * Calls notify on all observers.
	 */
	public void update() {
		// No support for observers yet
	}

	/**
	 * Dispose all instruction handlers of this list
	 */
	public void dispose() {
		InstructionHandle[] handles = getInstructionHandles();
		for (int i = handles.length - 1; i >= 0; i--) {
			InstructionHandle handle = handles[i];
			handle.dispose();
		}
		this.list.clear();
	}

	/**
	 * Appends the list in the compound instruction.
	 *
	 * @param arg0
	 * @return
	 */
	public InstructionHandle append(CompoundInstruction ci) {
		return this.append(ci.getInstructionList());
	}

	/**
	 * Inserts instruction i right after instruction handle ih. Returns
	 * a handle to i.
	 *
	 * @param ih
	 * @param i
	 * @return
	 */
	public InstructionHandle append(InstructionHandle ih, Instruction i) {
		InstructionHandle handle = i.createHandle();
		insertAfter(ih, handle);
        handle.list = this;
		return handle;
	}

	/**
	 * Sets the positions of the instructions.
	 */
	public void setPositions() {
		this.setPositions(false);
	}

	/**
	 * The length of the list.
	 * 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public int getLength() {
		int count = 0;
		ListIterator<AbstractInsnNode> iterator = list.iterator();
		while(iterator.hasNext())
			if (isBCELInstuction(iterator.next()))
				count++;
		return count;
	}

	/**
	 * Inserts the instruction list returned by this compound instruction.
	 *
	 * @param compoundInstruction
	 * @return
	 */
	public InstructionHandle insert(CompoundInstruction compoundInstruction) {
		if (this.list.size() == 0) {
			return this.append(compoundInstruction.getInstructionList());
		} else {
			return this.append(null, compoundInstruction);
		}
	}

	/**
	 * Appends this instruction to the end of the list. Returns a handle to the
	 * added instruction.
	 * 
	 * @param instruction
	 * @return
	 */
	public InstructionHandle append(Instruction instruction) {
		InstructionHandle handle = instruction.createHandle();
		append(handle);
		return handle;
	}

	/**
	 * Inserts the list in the compound instruction after this instruction handle. 
	 *
	 * @param ih
	 * @param i
	 * @return
	 */
	public InstructionHandle append(InstructionHandle ih, CompoundInstruction i) {
		InstructionHandle ret = null, prev = ih;
		
		for (InstructionHandle h : i.getInstructionList().getInstructionHandles()) {
			prev = this.append(prev, h.instruction);
			if (null == ret)
				ret = prev;
		}
		
		return ret;
	}

	/**
	 * Appends all instructions in the parameter list to the end of
	 * this list. Returns a handle to the first appended element.
	 *
	 * @param list
	 * @return
	 */
	public InstructionHandle append(InstructionList list) {
		InstructionHandle ih = null; //WrapperRecycler.wrap(list.list.getFirst(), InstructionHandle.class);
		for(InstructionHandle handle : list.getInstructionHandles()) {
			this.append(handle);
			if (null == ih)
				ih = handle;
		}
		return ih;
	}

	/**
	 * Inserts the branch instruction before this instruction handle. Returns a handle to the
	 * branch instruction.
	 *
	 * @param ih
	 * @param bi
	 * @return
	 */
	public BranchHandle insert(InstructionHandle ih, BranchInstruction bi) {
		BranchHandle bh = (BranchHandle) bi.createHandle();
		insertBefore(ih, bh);
        bh.list = this;
		return bh;
	}

	/**
	 * Same as length.
	 * 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public int size() {
		int i = 0;
		ListIterator<AbstractInsnNode> iterator = list.iterator();
		while(iterator.hasNext())
			if (isBCELInstuction(iterator.next()))
				i++;
		return i;
	}

	/**
	 * Appends to the end of the list.
	 *
	 * @param bi
	 * @return
	 */
	public BranchHandle append(BranchInstruction bi) {
		BranchHandle bh = (BranchHandle) bi.createHandle();
        append(bh);
        return bh;
	}

	/**
	 * Sets the position index in the instruction handlers of this list.
	 *
	 * @param check
	 */
	public void setPositions(boolean check) {
		int pos = 0;
		for (InstructionHandle ih : getInstructionHandles()) {
			ih.setPosition(pos);
			pos += InstructionHandle.getByteCount(ih.getNode());
		}
	}

	/**
	 * Checks whether the size of the list is zero.
	 * 
	 * @return
	 */
	public boolean isEmpty() {
		return this.list.size() == 0;
	}

	/**
	 * Gets a handle to the start instruction node.
	 * 
	 * @return
	 */
	public InstructionHandle getStart() {
		return this.getNext(null);
	}

	/**
	 * Gets a handle to the end instruction node.
	 * 
	 * @return
	 */
	public InstructionHandle getEnd() {
		return this.getPrev(null);
	}
	
	/**
	 * Return an array with the instruction handles that hold the ASM
	 * nodes for this list.
	 *
	 * @return
	 */
	public InstructionHandle[] getInstructionHandles() {
		List<InstructionHandle> handles = new ArrayList<InstructionHandle>();
		ListIterator<?> iterator = this.list.iterator();
		while(iterator.hasNext()) {
			AbstractInsnNode node = (AbstractInsnNode) iterator.next();
			if (isBCELInstuction(node))
				handles.add(WrapperRecycler.wrap(node, InstructionHandle.class));
		}
		return handles.toArray(new InstructionHandle[handles.size()]);
	}
	
	/**
	 * Returns an array with clones of the instructions of this list.
	 *
	 * @return
	 */
	public Instruction[] getInstructions() {
		this.setPositions();
		List<Instruction> instructions = new ArrayList<Instruction>();
		ListIterator<?> iterator = this.list.iterator();
		while(iterator.hasNext()) {
			AbstractInsnNode node = (AbstractInsnNode) iterator.next();
			if (! (node instanceof LabelNode))
				instructions.add(WrapperRecycler.wrap(node, InstructionHandle.class).getInstruction().clone());
		}
		return instructions.toArray(new Instruction[instructions.size()]);
	}
	
	/**
	 * Removes the handler from the underlying list.
	 * 
	 * Perhaps there's more cleanup needed.
	 *
	 * @param ih
	 */
	public void delete(InstructionHandle ih) {
		if (ih.hasTargeters())
			for (InstructionTargeter targeter : ih.getTargeters())
				if (targeter instanceof CodeExceptionGen)
					((CodeExceptionGen) targeter).replace(ih);
		
		this.list.remove(ih.getNode());
		ih.dispose();
	}
	
	// FROM HERE ON, WRAPPER MEMBERS
	
	/**
	 * Inserts ih after pos.
	 *
	 * @param pos
	 * @param ih
	 */
	@WrapperMember
	void insertAfter(InstructionHandle pos, InstructionHandle ih) {
		
		if (null == pos) {
			this.list.insert(ih.getNode());
			return;
		}
		
        this.list.insert(pos.getNode(), ih.getNode());
	}
	
	/**
	 * Inserts the instruction handle ih before the instruction pos.
	 *
	 * A better description is that ih is inserted after the previous node
	 * of pos. That's because there may be non-BCEL instructions (line numbers,
	 * labels, frame nodes) before pos, and they should be kept after ih.
	 * 
	 * @param pos
	 * @param ih
	 */
	@WrapperMember
	void insertBefore(InstructionHandle pos, InstructionHandle ih) {
		
		if (null == pos) {
			this.list.insert(ih.getNode());
			return;
		}
			
		// this ensures that the label, line and frame nodes will be kept before the handle and after its prev
		InstructionHandle prev = getPrev(pos.instruction.getNode());
		if (null == prev)
			prev = pos;
			
        this.list.insert(prev.getNode(), ih.getNode());
	}
	
	/**
	 * Helper method to append the handle to the end of this list.
	 *
	 * @param handle
	 */
	@WrapperMember
	void append(InstructionHandle handle) {
		this.list.add(handle.getNode());
		handle.list = this;
	}
	
	/**
	 * Checks whether this instruction handle already
	 * contains a corresponding LabelNode in its
	 * ASM representation.
	 *
	 * @param ih
	 * @return true, if this handle already has a corresponding label.
	 */
	@WrapperMember
	public boolean hasLabel(InstructionHandle ih) {
		return null != getLabel(ih);
	}
	
	/**
	 * Checks whether this instruction handle already
	 * contains a corresponding next LabelNode in its
	 * ASM representation.
	 *
	 * @param ih
	 * @return true, if this handle already has a corresponding next label.
	 */
	@WrapperMember
	public boolean hasNextLabel(InstructionHandle ih) {
		return null != getNextLabel(ih);
	}
	
	/**
	 * Gets the label corresponding to this instruction
	 * handle, if it exists.
	 *
	 * @param ih
	 * @return the corresponding label or null.
	 */
	@WrapperMember
	public LabelNode getLabel(InstructionHandle ih) {
		AbstractInsnNode node = ih.getNode();
		if (! list.contains(node))
			return null;
		
		node = node.getPrevious();
		while(node != null && ! isBCELInstuction(node) && ! (node instanceof LabelNode))
			node = node.getPrevious();
		
		if (node instanceof LabelNode)
			return (LabelNode) node;
		return null;
	}
	
	/**
	 * Gets the next label corresponding to this instruction
	 * handle, if it exists.
	 *
	 * @param ih
	 * @return the corresponding label or null.
	 */
	@WrapperMember
	public LabelNode getNextLabel(InstructionHandle ih) {
		InstructionHandle next = getNext(ih.getNode());
		if (null != next)
			return getLabel(next);
		
		// ih is the last instruction
		AbstractInsnNode label = ih.getNode();
		while(label != null && ! (label instanceof LabelNode))
			label = label.getNext();

		if (label instanceof LabelNode)
			return (LabelNode) label;
		return null;
	}
	
	/**
	 * Locates the line number node that refers to this handle, if
	 * it exists.
	 *
	 * @param ih
	 * @return
	 */
	@WrapperMember
	public LineNumberNode getLineNumber(InstructionHandle ih) {
		AbstractInsnNode node = ih.getNode();
		if (! list.contains(node))
			return null;
		
		node = node.getPrevious();
		while(node != null && ! (node instanceof LineNumberNode))
			node = node.getPrevious();
		
		if (node instanceof LineNumberNode)
			return (LineNumberNode) node;
		return null;
	}
	
	/**
	 * Makes sure that this handle has a corresponding label
	 * in the list. This method will return the existing label
	 * if it exists, or will create a new one if necessary.
	 *
	 * @param ih
	 * @return
	 */
	@WrapperMember
	public LabelNode ensureLabel(InstructionHandle ih) {
		LabelNode label = getLabel(ih);
		if (null == label) {
			label = new LabelNode();
			this.list.insertBefore(ih.getNode(), label);
		}
		return label;
	}
	
	/**
	 * Makes sure that this handle has a corresponding "next"
	 * label. That is, there must be a label after this instruction
	 * but before the next one (or the end of the list).
	 *
	 * @param ih
	 * @return
	 */
	@WrapperMember
	public LabelNode ensureNextLabel(InstructionHandle ih) {
		InstructionHandle next = getNext(ih.getNode());
		if (null != next)
			return ensureLabel(next);
		
		// ih is the last instruction
		AbstractInsnNode label = ih.getNode();
		while(label != null && ! (label instanceof LabelNode))
			label = label.getNext();
		
		// add a new label at the end of the list
		if (null == label) {
			label = new LabelNode();
			this.list.add(label);
		}
		return (LabelNode) label;
	}
	
	/**
	 * Gets the handle that corresponds to this label.
	 *
	 * @param label
	 * @return
	 */
	@WrapperMember
	public InstructionHandle getHandle(LabelNode label) {
		// TODO - we could optimize this with a cache
		return getNext(label);
	}
	
	/**
	 * Gets the previous handle with respect to this label.
	 *
	 * @param label
	 * @return
	 */
	@WrapperMember
	public InstructionHandle getPrevHandle(LabelNode label) {
		// TODO - we could optimize this with a cache
		return getPrev(label);
	}
	
	/**
	 * Gets a handle to the next instruction in the list, or null if this is the last instruction.
	 *
	 * If node is null, this method will return the first instruction handle of the list.
	 * 
	 * Note that only BCEL instructions correspond to InstructionHandles.
	 * 
	 * @return
	 */
	@WrapperMember
	public final InstructionHandle getNext(AbstractInsnNode node) {
		
		AbstractInsnNode next = (null == node) ? list.getFirst() : node.getNext();
		while(null != next && ! isBCELInstuction(next))
			next = next.getNext();
		
		if (null == next)
			return null;
		
		return WrapperRecycler.wrap(next, InstructionHandle.class);
	}

	/**
	 * Gets a handle to the previous instruction in the list, or null if this is the first instruction. 
	 *
	 * If node is null, this method will return the last instruction handle of the list.
	 * 
	 * @return
	 */
	@WrapperMember
	public final InstructionHandle getPrev(AbstractInsnNode node) {
		AbstractInsnNode prev = (null == node) ? list.getLast() : node.getPrevious();
		while(null != prev && ! isBCELInstuction(prev))
			prev = prev.getPrevious();
		
		if (null == prev)
			return null;
		
		return WrapperRecycler.wrap(prev, InstructionHandle.class);
	}
	
	/**
	 * Checks whether this node corresponds to an instruction in BCEL.
	 * 
	 * Label, Frame and LineNumber nodes are present in ASM instruction lists
	 * but are not instructions in BCEL.
	 *
	 * @param node
	 * @return
	 */
	@WrapperMember
	public static boolean isBCELInstuction(AbstractInsnNode node) {
		return ! (node instanceof LabelNode || node instanceof FrameNode || node instanceof LineNumberNode);
	}
	
}
