/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import java.util.List;
import java.util.Map;

import org.apache.bcel.support.PushedDown;
import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LookupSwitchInsnNode;

/**
 * 
 * TODO Comment!!
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public class LOOKUPSWITCH extends Select {

	public LOOKUPSWITCH(int[] match, InstructionHandle[] targets, InstructionHandle defaultTarget) {
		// Get the labels for each target
		LabelNode[] labels = new LabelNode[targets.length];
		for (int i = 0; i < labels.length; i++)
			labels[i] = targets[i].ensureLabel();
		
		// Create the wrappee
		setNode(new LookupSwitchInsnNode(
					defaultTarget.ensureLabel(),
					match,
					labels
				));
		
		// Now set the targeters
		for (int i = 0; i < labels.length; i++)
			registerTargeter(targets[i]);
		
		// The default target becomes this instruction's "main target", so register last
		registerTargeter(defaultTarget);
	}
	
	/**
	 * Computes the list of targets of this switch.
	 *
	 * @return
	 */
	@SuppressWarnings("unchecked")
	@PushedDown
	public InstructionHandle[] getTargets() {
		
		InstructionHandle handle = WrapperRecycler.wrap(getNode(), InstructionHandle.class);
		
		List<LabelNode> labels = ((LookupSwitchInsnNode) getNode()).labels;
		InstructionHandle[] targets = new InstructionHandle[labels.size()];
		for (int i = 0; i < targets.length; i++)
			targets[i] = handle.list.getHandle(labels.get(i));
		
		return targets;
	}
	
	@SuppressWarnings("unchecked")
	@PushedDown
	public void setTarget(int index, InstructionHandle newTarget) {
		InstructionHandle ih = getHandleFromList(newTarget.list);		
		assert(ih != null);
		
		// Replace the label
		LabelNode oldTarget = (LabelNode) ((LookupSwitchInsnNode) ih.getNode()).labels.set(index, newTarget.ensureLabel());
		
		// De-register as a targeter from old handler, register to new
		ih.list.getHandle(oldTarget).removeTargeter(this);
		newTarget.addTargeter(this);
	}
	
	@WrapperMember
	LOOKUPSWITCH() {
	}
	
	/**
	 * Getter allows for wrappee refinement.
	 *
	 * @return
	 */
	@WrapperMember
	protected LookupSwitchInsnNode getNode() {
		return (LookupSwitchInsnNode) super.getNode();
	}
	
	/**
	 * Gets the label that is the target of this branch.
	 *
	 * @param label
	 */
	@WrapperMember
	LabelNode getTargetLabel() {
		return getNode().dflt;
	}
	
	/**
	 * Sets the label that is the target of this branch.
	 *
	 * @param label
	 */
	@WrapperMember
	void setTargetLabel(AbstractInsnNode node, LabelNode label) {
		((LookupSwitchInsnNode) node).dflt = label;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	@WrapperMember
	Instruction rewrite(AbstractInsnNode node, Map<LabelNode, InstructionHandle> map) {
		LookupSwitchInsnNode i = (LookupSwitchInsnNode) node;

		// Keep the default label because it is always overwritten by registerTargeter
		LabelNode defaultLabel = i.dflt;
		
		// Compute the list of new target labels and set to the instruction
		InstructionHandle[] targets = new InstructionHandle[i.labels.size()];
		for (int j = 0; j < targets.length; j++)
			targets[j] = map.get(i.labels.get(j));
		
		i.labels.clear();
		for (int j = 0; j < targets.length; j++) {
			i.labels.add(targets[j].ensureLabel());
			registerTargeter(targets[j]);
		}
			
		// Compute the new default target label
		InstructionHandle defaultTarget = map.get(defaultLabel);
		i.dflt = defaultTarget.ensureLabel();
		registerTargeter(defaultTarget);
		return this;
	}
	
	@Override
	@WrapperMember
	public Instruction cloneInstruction() {
		LOOKUPSWITCH i = new LOOKUPSWITCH();
		i.setNode(this.cloneNode());
		return i;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	@WrapperMember
	public AbstractInsnNode cloneNode() {
		LookupSwitchInsnNode n = (LookupSwitchInsnNode) getNode();
		int[] keys = new int[n.keys.size()];
		for (int i = 0; i < keys.length; i++)
			keys[i] = (Integer) n.keys.get(i);
		
		return new LookupSwitchInsnNode(
				n.dflt,
				keys,
				(LabelNode[]) n.labels.toArray(new LabelNode[n.labels.size()]));
	}
	
	
	/**
	 * Returns all labels that are targets of this instruction.
	 *
	 * @return
	 */
	@WrapperMember
	public LabelNode[] getTargetLabels() {
		LookupSwitchInsnNode n = (LookupSwitchInsnNode) getNode();
		LabelNode[] targets = new LabelNode[n.labels.size() + 1];
		
		targets[0] = n.dflt;
		for (int i = 1; i < targets.length; i++)
			targets[i] = (LabelNode) n.labels.get(i - 1);
		
		return targets;
	}
}

