/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.support.PushedDown;
import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.FrameNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LineNumberNode;
import org.objectweb.asm.tree.LocalVariableNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.TryCatchBlockNode;

/**
 * 
 * TODO Comment!!
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public class MethodGen extends FieldGenOrMethodGen {
	
	/**
	 * Stateful adapter.
	 */
	@WrapperMember
	String className = null;
	
	/**
	 * Getter allows for wrappee refinement.
	 *
	 * @return
	 */
	@WrapperMember
	protected MethodNode getMemberNode() {
		return (MethodNode) super.getMemberNode();
	}
	
	/**
	 * 
	 * TODO Comment!!
	 * 
	 * @param access_flags
	 * @param return_type
	 * @param arg_types
	 * @param arg_names
	 * @param method_name
	 * @param class_name
	 * @param il
	 * @param cp
	 */
	public MethodGen(int access_flags, Type return_type, Type[] arg_types, String[] arg_names, String method_name, String class_name, InstructionList il, ConstantPoolGen cp) {

		this.className = class_name;
		
		// init with a new method node
		MethodNode node = new MethodNode();
		node.access = access_flags;
		node.desc = org.objectweb.asm.Type.getMethodDescriptor(return_type.getType(), Type.convertTypes(arg_types));
		node.name = method_name;
		node.exceptions = new ArrayList<String>(0);
        
		this.initInstructionList(node, il, arg_types, arg_names);
		this.setMemberNode(node);
	}
	
	/**
	 * Initializes this instruction list with local variables and set it to this node.
	 *
	 * @param node
	 * @param il
	 * @param arg_types
	 * @param arg_names
	 */
	@WrapperMember
	@SuppressWarnings("unchecked")
	void initInstructionList(MethodNode node, InstructionList il, Type[] arg_types, String[] arg_names) {
		int access_flags = node.access;
		
		// create "this" and parameter variables
		if ((Opcodes.ACC_ABSTRACT & access_flags) == 0) {
			int index = 0;
			InstructionHandle start = il.getStart();
			InstructionHandle end = il.getEnd();
			
			node.localVariables = new ArrayList<LocalVariableNode>();
			// "this" only if not static and not abstract
			if ((Opcodes.ACC_STATIC & access_flags) == 0)
				node.localVariables.add(new LocalVariableGen(index++, "this", new ObjectType(this.className), start, end).getNode());
			
			// parameters only if not abstract
			for (int i = 0; i < arg_types.length; i++) {
				node.localVariables.add(new LocalVariableGen(index++, arg_names[i], arg_types[i], start, end).getNode());
			}
		}
		node.instructions = il.list;
	}
	
	
	/**
	 * Creates a MethodGen using this Method.
	 *
	 * @param m
	 * @param class_name
	 * @param cp
	 */
	public MethodGen(Method prototype, String class_name, ConstantPoolGen cp) {
		this.className = class_name;
		this.setMemberNode(clone(prototype.getNode(), true));
	}

	
	/**
	 * Creates a new method node which is a clone of this orig node.
	 * 
	 * The cloned node contains the basic information of the original node,
	 * like access flags, names and throws declaration, but does not yet
	 * have instructions, exception handlers or local variables.
	 *
	 * @param orig
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static MethodNode createClone(MethodNode orig) {
		
		// clone the method node
		MethodNode node = new MethodNode();
		node.access = orig.access;
		node.desc = orig.desc;
		node.name = orig.name;
		node.exceptions = new ArrayList<String>(0);
		node.exceptions.addAll(orig.exceptions);
		
		return node;
	}
		
	/**
	 * Creates an unpacked clone of the orig node, which is assumed to be packed.
	 *
	 * @param orig
	 * @param keepLocalVariableTargeters
	 * @return
	 */
	@SuppressWarnings("unchecked")
	@WrapperMember
	public static MethodNode clone(MethodNode orig, boolean keepLocalVariableTargeters) {
		
		// 1) basic cloning
		MethodNode clone = createClone(orig);
		InstructionList cloneList = new InstructionList(clone.instructions);
		
		// 2) clone labels
		Map<LabelNode, LabelNode> clonedLabels = cloneLabels(orig.instructions);
		
		// 3) Clone instructions at ASM level and already create handles
		ListIterator<AbstractInsnNode> iterator = orig.instructions.iterator(); 
		while(iterator.hasNext()) {
			AbstractInsnNode i = iterator.next();

			if (InstructionList.isBCELInstuction(i)) {
				AbstractInsnNode newI = i.clone(clonedLabels);
				clone.instructions.add(newI);
				
				InstructionHandle handle = WrapperRecycler.wrap(i, InstructionHandle.class);
				InstructionHandle newHandle = handle.getInstruction().createHandle(newI);
				newHandle.list = cloneList;
				
				// Local variable gens are sometimes kept. This looks like a bug in BCEL
				/*
				 * 
				 * 
				if (keepLocalVariableTargeters && handle.hasTargeters()) {
					for (InstructionTargeter t : handle.getTargeters())
						if (t instanceof LocalVariableGen && "this".equals(((LocalVariableGen) t).getNode().name))
							newHandle.addTargeter(t);
				}*/
			} else if (i instanceof LabelNode) {
				clone.instructions.add(clonedLabels.get(i));
				
			} else {
				clone.instructions.add(i.clone(clonedLabels));
			}
		}
		
		// 4) Exception handlers
		if (null != orig.tryCatchBlocks)
			cloneExceptionHandlers(orig, clone, clonedLabels);
		
		// 5) Local variables
		if (null != orig.localVariables)
			cloneLocalVariables(orig, clone, clonedLabels);
		
		// 6) Set targeter relations for BCEL instructions that are targeters
		for (InstructionHandle handle : cloneList.getInstructionHandles()) {
			InstructionTargeter targeter = handle.getInstructionTargeter();
			if (null != targeter)
				for (LabelNode l : targeter.getTargetLabels())
					handle.list.getHandle(l).addTargeter(targeter);
		}

		// 7) Set targeter relations for LineNumberGen
		iterator = clone.instructions.iterator(); 
		while(iterator.hasNext()) {
			AbstractInsnNode i = iterator.next();
			 if (i instanceof LineNumberNode) {
				 LineNumberNode line = (LineNumberNode) i;
				 cloneList.getHandle(line.start).addTargeter(new LineNumberGen(line));
			 }
		}
		
		// 8) make sure positions are calculated on the cloned list
		cloneList.setPositions();
		
		return clone;
	}
	
	
	@SuppressWarnings("unchecked")
	public static void cloneLocalVariables(MethodNode orig, MethodNode clone, Map<LabelNode, LabelNode> clonedLabels) {
		if (null != orig.localVariables) {
			List<LocalVariableNode> clones = new ArrayList<LocalVariableNode>();
			
			ListIterator<LocalVariableNode> lvIterator = orig.localVariables.listIterator(); 
			while(lvIterator.hasNext()) {
				LocalVariableNode node = lvIterator.next();
				LocalVariableNode lvnClone = 
						new LocalVariableNode(
								node.name, 
								node.desc, 
								node.signature, 
								clonedLabels.get(node.start), 
								clonedLabels.get(node.end),
								node.index);
				
				LocalVariableGen gen = WrapperRecycler.wrap(lvnClone, LocalVariableGen.class);
				InstructionList list = WrapperRecycler.wrap(clone.instructions, InstructionList.class);
				
				list.getHandle(lvnClone.start).addTargeter(gen);
				InstructionHandle h = list.getHandle(lvnClone.end);
				// adjust if the end is the last label of the list
				if (null == h)
					h = list.getPrevHandle(lvnClone.end);
				h.addTargeter(gen);
				
				clones.add(lvnClone);
			}
			
			clone.localVariables = clones;
		}
	}
	
	@SuppressWarnings("unchecked")
	public static void cloneExceptionHandlers(MethodNode orig, MethodNode clone, Map<LabelNode, LabelNode> clonedLabels) {
		if (null != orig.tryCatchBlocks) {
			List<TryCatchBlockNode> tcbNodes = new ArrayList<TryCatchBlockNode>();
		
			ListIterator<TryCatchBlockNode> tcbIterator = orig.tryCatchBlocks.listIterator(); 
			while(tcbIterator.hasNext()) {
				TryCatchBlockNode origTCB = tcbIterator.next();
				CodeExceptionGen origGen = WrapperRecycler.wrap(origTCB, CodeExceptionGen.class);
				
				TryCatchBlockNode cloneTCB =
						new TryCatchBlockNode(
								clonedLabels.get(origTCB.start), 
								clonedLabels.get(origTCB.end), 
								clonedLabels.get(origTCB.handler),
								origTCB.type);
				
				CodeExceptionGen gen = WrapperRecycler.wrap(cloneTCB, CodeExceptionGen.class);
				gen.type = new ObjectType(origGen.type.getClassName());
				gen.list = WrapperRecycler.wrap(clone.instructions, InstructionList.class);
				
				gen.list.getHandle(cloneTCB.start).addTargeter(gen);
				gen.list.getHandle(cloneTCB.handler).addTargeter(gen);
				gen.list.getPrevHandle(cloneTCB.end).addTargeter(gen);
				
				tcbNodes.add(cloneTCB);
			}
			
			clone.tryCatchBlocks = tcbNodes;
		}
	}
	/**
	 * Returns a map from labels present in this list to a brand new clone.
	 *
	 * @param list
	 * @return
	 */
	@SuppressWarnings("unchecked")
	@WrapperMember
	public static Map<LabelNode, LabelNode> cloneLabels(InsnList list) {
		Map<LabelNode, LabelNode> labels = new HashMap<LabelNode, LabelNode>();
		ListIterator<AbstractInsnNode> iterator = list.iterator(); 
		while(iterator.hasNext()) {
			AbstractInsnNode node = iterator.next();
			if (node instanceof LabelNode)
				labels.put((LabelNode) node, new LabelNode());
		}
		return labels;
	}
	
	/**
	 * Constructs an array with the names of local variables in this method node,
	 * which come from parameters (i.e., not "this").
	 * 
	 * TODO - this will also include local variables, we have to restrict it!
	 *
	 * @param node
	 * @return
	 */
	@WrapperMember
	public static String[] getParameterNames(MethodNode node) {
		List<String> names = new ArrayList<String>();
		for (Object o : node.localVariables) {
			LocalVariableNode local = (LocalVariableNode) o;
			if (! ("this".equals(local.name)))
				names.add(local.name);
		}
		return names.toArray(new String[names.size()]);
	}
	
	@WrapperMember
	public static Type[] getParameterTypes(MethodNode node) {
		return Type.convertTypes(org.objectweb.asm.Type.getArgumentTypes(node.desc));
	}
	
	/**
	 * Gets the name of the class that contains this method.
	 *
	 * @return
	 */
	public String getClassName() {
		return className;
	}
	
	/**
	 * Gets a line number table for the method.
	 *
	 * @param arg0
	 * @return
	 */
	public LineNumberTable getLineNumberTable(ConstantPoolGen cp) {
		return new LineNumberTable(this.getInstructionList());
	}

	/**
	 * Computes the max stack size.
	 */
	public void setMaxStack() {
		// ASM automatically computes the max stack
	}

	/**
	 * Removes NOP nodes from the list.
	 */
	public void removeNOPs() {
		InstructionList list = WrapperRecycler.wrap(getMemberNode().instructions, InstructionList.class);
		for (InstructionHandle h : list.getInstructionHandles())
			if (Opcodes.NOP == h.getNode().getOpcode())
				list.delete(h);
	}

	/**
	 * Computes the max number of locals.
	 */
	public void setMaxLocals() {
		// ASM automatically computes the max locals
	}

	/**
	 * Returns the instruction list of this method.
	 * 
	 * @return
	 */
	public InstructionList getInstructionList() {
		return WrapperRecycler.wrap(getMemberNode().instructions, InstructionList.class);
	}

	/**
	 * Sets a new instruction list for this method.
	 *
	 * @param list
	 */
	@SuppressWarnings("unchecked")
	public void setInstructionList(InstructionList list) {
		
		// In BCEL line numbers are a property of the method, while in ASM they are
		// nodes in the instruction list. When a new list is set, we must translate
		// the line number nodes from the old list to the new one.
		
		// First replace the list and initialize the new list, making sure positions are calculated
		MethodNode methodNode = getMemberNode();
		InsnList oldList = methodNode.instructions;
		InsnList newList = list.list;
		list.setPositions();
		
		methodNode.instructions = newList;
		
		// Create a mapping between labels from the old list to labels in the new list
		Map<LabelNode, LabelNode> map = new HashMap<LabelNode, LabelNode>();
		InstructionList oldIList = WrapperRecycler.wrap(oldList, InstructionList.class);

		InstructionHandle o = oldIList.getStart();
		InstructionHandle n = list.getStart();
		
		int nops = 0;
		while(null != o) {
			// ignore NOPs when matching labels
			while(n != null && n.instruction.getOpcode() == Opcodes.NOP) {
				n = n.getNext();
				nops++;
			}
			
			if (null == n)
				break;
			
			// found correspondence between original and new labels
			if (o.hasLabel())
				map.put(o.ensureLabel(), n.ensureLabel());
			
			o = o.getNext();
			
			// if o ran over and n will also run over, match them with "ensureNextLabel"
			if (null == o && null == n.getNext())
				n = null;
			
			// inc n to match the position of o
			while(n != null && o != null && o.getPosition() + nops > n.getPosition())
				n = n.getNext();
		}
		
		if (oldIList.getEnd().hasNextLabel()) {
			if (null == n) {
				map.put(oldIList.getEnd().ensureNextLabel(), list.getEnd().ensureNextLabel());
			} else {
				map.put(oldIList.getEnd().ensureNextLabel(), n.ensureLabel());
			}
		}
		
		// Now set the line number and frame nodes based on the labels
		ListIterator<AbstractInsnNode> iterator = oldList.iterator();
		LineNumberNode clone = null;
		while(iterator.hasNext()) {
			AbstractInsnNode ln = iterator.next();
			if (ln instanceof LineNumberNode) {
				// insert the line number node right after its label
				clone = (LineNumberNode) ln.clone(map);
				newList.insert(clone.start, clone);
				
				
			} else if (ln instanceof FrameNode) {
				newList.insert(clone, ln.clone(map));
			}
		}
		
		// Now we can clone the handlers and local variables using the label map
		cloneExceptionHandlers(methodNode, methodNode, map);
		cloneLocalVariables(methodNode, methodNode, map);
	}

	/**
	 * Update observers of this method.
	 * 
	 * Currently nothing is done.
	 *
	 */
	public void update() {
		// Nothing to be done because we do not use observers
	}

	/**
	 * Adds a new throws declaration to this method.
	 *
	 * @param exceptionName
	 */
	@SuppressWarnings("unchecked")
	public void addException(String exceptionName) {
		MethodNode node = this.getMemberNode();
		if (null == node.exceptions)
			node.exceptions = new ArrayList<String>();
		node.exceptions.add(Type.getInternalName(exceptionName));
	}

	/**
	 * Gets the generated method object.
	 * 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public Method getMethod() {
		
		// Initialize local variable labels if needed
		MethodNode orig = getMemberNode();
		
		if (orig.localVariables.size() > 0) {
			LabelNode start = null; 
			LabelNode end = null;
			
			ListIterator<LocalVariableNode> i = orig.localVariables.listIterator();
			while(i.hasNext()) {
				LocalVariableNode lvn = i.next();
				if (null == lvn.start) {
					if (null == start)
						start = this.getInstructionList().getStart().ensureLabel();
					lvn.start = start;
				}
				if (null == lvn.end) {
					if (null == end)
						end = this.getInstructionList().getEnd().ensureNextLabel();
					lvn.end = end;
				}
			}
		}
			
		// Clone and initialize the node
		MethodNode n = clone(orig, false);

		// using BCEL's code to order the local variables
		if (n.localVariables.size() > 0)
			sort(n.localVariables, 0, n.localVariables.size() - 1);
		
		// Recompute positions also on this methodgen
		this.getInstructionList().setPositions();
		
		// Always creates a new method
		return new Method(n);
	}

	/**
	 * This is a clone of BCEL's code to sort local variables, with minor changes
	 * to use LocalVariableNodes.
	 *
	 * We have submitted an enhancement proposal to BCEL, to remove this code and
	 * use Java's own sort implementation: 
	 * 
	 * https://issues.apache.org/bugzilla/show_bug.cgi?id=52433
	 * 
	 * @param vars
	 * @param l
	 * @param r
	 */
	private static final void sort(List<LocalVariableNode> vars, int l, int r ) {
        int i = l, j = r;
        int m = vars.get((l + r) / 2).index;
        LocalVariableNode h;
        do {
            while (vars.get(i).index < m) {
                i++;
            }
            while (m < vars.get(j).index) {
                j--;
            }
            if (i <= j) {
                h = vars.get(i);
                vars.set(i, vars.get(j));
                vars.set(j, h); // Swap elements
                i++;
                j--;
            }
        } while (i <= j);
        if (l < j) {
            sort(vars, l, j);
        }
        if (i < r) {
            sort(vars, i, r);
        }
    }
	
	/**
	 * Adds a new exception handler to this method's code.
	 *
	 * @param start_pc
	 * @param end_pc
	 * @param handler_pc
	 * @param catch_type
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public CodeExceptionGen addExceptionHandler(InstructionHandle start_pc, InstructionHandle end_pc, InstructionHandle handler_pc, ObjectType catch_type) {
		CodeExceptionGen ceg = new CodeExceptionGen(start_pc, end_pc, handler_pc, catch_type);
		MethodNode node = this.getMemberNode();
		if (null == node.tryCatchBlocks)
			node.tryCatchBlocks = new ArrayList<TryCatchBlockNode>();

		node.tryCatchBlocks.add(ceg.getNode());
		return ceg;
	}

	/**
	 * Gets the exception handlers declared for this method.
	 * 
	 * @return
	 */
	public CodeExceptionGen[] getExceptionHandlers() {
		List<?> blocks = getMemberNode().tryCatchBlocks;
		if (null == blocks || blocks.isEmpty())
			return new CodeExceptionGen[0];
		
		CodeExceptionGen[] array = new CodeExceptionGen[blocks.size()];
		for (int i = 0; i < array.length; i++)
			array[i] = WrapperRecycler.wrap(blocks.get(i), CodeExceptionGen.class);
		return array;
	}

	/**
	 * Remove all exception handlers from this method.
	 */
	public void removeExceptionHandlers() {
		List<?> blocks = this.getMemberNode().tryCatchBlocks;
		if (null != blocks)
			blocks.clear();
	}
	
	/**
	 * Originally from super type.
	 * 
	 * @return
	 */
	@Override
	@PushedDown
	public String getName() {
		return getMemberNode().name;
	}

	/**
	 * Originally on super type.
	 * 
	 * @return
	 */
	@Override
	@PushedDown
	public boolean isPublic() {
		throw new UnsupportedOperationException();
	}

	/**
	 * Originally on super type.
	 * 
	 * @param arg0
	 */
	@Override
	@PushedDown
	public void setAccessFlags(int arg0) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Originally on super type.
	 * 
	 * @return
	 */
	@Override
	@PushedDown
	public int getAccessFlags() {
		throw new UnsupportedOperationException();
	}

	/**
	 * TODO Comment!!
	 *
	 * @param className2
	 * @param constantPool
	 * @return
	 */
	public MethodGen copy(String className2, ConstantPoolGen constantPool) {
		// TODO Auto-generated method stub
		return this;
	}
}
